#!/bin/bash

########################################################################
#   SaraB - Schedule And Rotate Automatic Backups
#   Copyright (C) 2004  Tristan R. Rhodes
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#   The author, Tristan Rhodes, can be contacted by email at
#      tristanbob@hotmail.com
########################################################################

########################################################################
########################################################################
#
#   Define functions that will be used later
#
########################################################################
########################################################################

## Rotates the rotation.schedule file by one line.  Top line moves to the bottom line.
function rotate 
{
    lines=$(cat $SARAB_ETC/$ROTATION_SCHEDULE | wc -l)
    firstline="$(head -n 1 $SARAB_ETC/rotation.schedule)"

    # Copy all but the first line back into rotation schedule
    tail -n $(expr $lines - 1) $SARAB_ETC/$ROTATION_SCHEDULE > $SARAB_ETC/rotation.schedule.temp

    # Append the first line to the end of rotation.schedule
    echo $firstline >> $SARAB_ETC/rotation.schedule.temp
    mv $SARAB_ETC/rotation.schedule.temp $SARAB_ETC/$ROTATION_SCHEDULE
}

## Finds the next virtual-tape information, rotating as needed.  Disregards all comments.
function get_nextarchive
{
    # Find the total lines in the file, so we can prevent infinite loops
    TOTAL_LINES=$(cat $SARAB_ETC/$ROTATION_SCHEDULE | wc -l)

    # Record the current first line in the file
    THIS_LINE=$(head -n 1 $SARAB_ETC/$ROTATION_SCHEDULE)

   # Get value of the first line, eliminating any comments
    DATA="$(echo $THIS_LINE | awk -F# '{print $1}')"
    COUNT=1
    
    # Loop until DATA is found or all lines have been processed  
    until [ $COUNT -gt $TOTAL_LINES ] || [ "$DATA" != "" ]; do  
          
	  # Rotate rotation.schedule by one line
          rotate
          if [ $? != 0 ]; then
             echo "ERROR: Error when rotating the rotation.schedule file."
             exit 1
          fi
  
          # Record the current first line in the file
          THIS_LINE=$(head -n 1 $SARAB_ETC/$ROTATION_SCHEDULE)

          # Look for data that is before any comments
	  DATA="$(echo $THIS_LINE | awk -F# '{print $1}')"

          # Increment line count 
          let COUNT+=1
	  
    done

    # Return the data. If none was found, this will return blank.
    echo $DATA
}

# Verbose output function
function verbose
{
    if [ "$SARAB_VERBOSE" = "yes" ]; then
       echo $1
    fi
}

function main
{

## Record start time
START_TIME=$(date +'%s')

# Verify input data
if [ "$DESTINATION" = "" ] || [ "$WORK_DIR" = "" ]; then
   echo "ERROR: You must specify the DESTINATION and WORK_DIR variables in sarab.conf"
   exit 1
fi

echo "****************************************************************************"
echo "Backup started at $(date)"
verbose "Reading settings from: $SARAB_CONF"
verbose "Using Dar configuration file named: $SARAB_ETC/$SARAB_DCF"
verbose "Using rotation.schedule named: $SARAB_ETC/$ROTATION_SCHEDULE"
verbose "Virtual-tapes are located under: $DESTINATION"
verbose "Log file located at: $LOG_FILE"
verbose

## If destination directory does not exist, create the directory
if [ ! -d "$DESTINATION/" ]; then
   verbose "Destination directory does not exist, creating directory... mkdir -p $DESTINATION/"
   mkdir -p $DESTINATION/
   if [ "$?" != "0" ]; then
      echo "ERROR: Error when creating the destination directory."
      exit 1
   fi
fi

## If working directory does not exist, create an empty directory to save the archive in
if [ ! -d "$DESTINATION/$WORK_DIR" ]; then
   verbose "Working directory does not exist, creating directory..."
   verbose "mkdir -p $DESTINATION/$WORK_DIR"
   mkdir -p $DESTINATION/$WORK_DIR
   if [ "$?" != "0" ]; then
      echo "ERROR: Error when creating the working directory."
      exit 1
   fi
else
   ## Remove files from working directory
   verbose "Removing temporary files..."
   verbose "rm -f $DESTINATION/$WORK_DIR/*"
   rm -f $DESTINATION/$WORK_DIR/*
   if [ "$?" != "0" ]; then
      echo "ERROR: Error when removing working directory files."
      exit 1
   fi
fi

## Record the current line from rotation.schedule
echo
echo "Reading the top line of the rotation.schedule file..."
CURRENT_LINE=$(get_nextarchive)
if [ "$?" != "0" ]; then
   echo "ERROR: Error when processing the rotation.schedule file."
   exit 1
fi

## Test if the rotation schedule has any uncommented data in it
if [ "$CURRENT_LINE" = "" ]; then
   echo "ERROR: The rotation.schedule file does not appear to have any data."
   exit 1
fi

verbose "Top line of rotation.schedule: $CURRENT_LINE"

## Record the number of fields in the current line, to see if it is a differential/incremental backup
NUM_FIELDS=$(echo $CURRENT_LINE | wc -w)
verbose "Number of fields: $NUM_FIELDS"

## If more than two fields, the data in rotation.schedule is not in the right format
if [ $NUM_FIELDS -gt 3 ]; then
   echo "Error: Each line in rotation.schedule can only have one, two or three fields."
   exit 1
fi

## Test to see if the user wants us to call the archive script on this backup
if [ $NUM_FIELDS -gt 1 ]; then
   LAST_WORD=$(echo $CURRENT_LINE | cut -f $NUM_FIELDS -d" ")
   if [ "$LAST_WORD" = "archive" ]; then
      CALL_ARCHIVE_SCRIPT="1"

      # Lets verify that the archive script to call actually exists...
      if [ -z $ARCHIVE_SCRIPT ]; then
         echo "Error: The archive script variable was not set in the config file"
       exit 1
      fi

      if [ ! -f $ARCHIVE_SCRIPT ]; then
         echo "Error: $ARCHIVE_SCRIPT does not exist"
       exit 1
      fi

      if [ ! -x $ARCHIVE_SCRIPT ]; then
               echo "Error: $ARCHIVE_SCRIPT is not executable"
       exit 1
      fi

      # Decrement our $NUM_FIELDS var for the other functions to work
      let NUM_FIELDS=$NUM_FIELDS-1 

      # Remove the archive from the line
      CURRENT_LINE="$(echo $CURRENT_LINE | awk -F" archive" '{ print $1 }')"

      echo "Calling archive script: yes"
   else
      echo "Calling archive script: no"
   fi
else 
   echo "Calling archive script: no"
fi


## Two fields means this is a diffential/incremental backup, one field means this is a full backup   
if [ $NUM_FIELDS = 2 ]; then
   # First field is the current archive
   CURRENT_ARCHIVE=$(echo $CURRENT_LINE | awk '{print $1}')
   # Second field is the reference archive
   REFERENCE_ARCHIVE=$(echo $CURRENT_LINE | awk '{print $2}')
   # Test to see if the reference archive actually exists
   if [ -d "$DESTINATION/$REFERENCE_ARCHIVE/" ]; then     # The reference archive exists  
   REFERENCE_BASENAME="--ref $(/bin/ls $DESTINATION/$REFERENCE_ARCHIVE/*.dar | head -n 1 | awk -F. '{print $1}')"
      # Record information about the reference archive to include in the current archive
      echo "The reference archive for this backup was:" > $DESTINATION/$WORK_DIR/reference_archive.txt
      echo "$(ls -ltr $DESTINATION/$REFERENCE_ARCHIVE/*.dar)" >> $DESTINATION/$WORK_DIR/reference_archive.txt
   else                              # The reference archive does not exist
      REFERENCE_BASENAME=""          # Do not use a reference archive, do a full-backup
   fi
else                                 # Only one field, do a full-backup
   CURRENT_ARCHIVE=$CURRENT_LINE
   REFERENCE_ARCHIVE=""
   REFERENCE_BASENAME=""
fi

# Output to the user what type of backup is being made, and to which virtual-tapes
if [ "$REFERENCE_ARCHIVE" = "" ]; then   # The user wants a full-backup
   echo "Creating a full-backup on the virtual-tape named: $CURRENT_ARCHIVE"
elif [ "$REFERENCE_BASENAME" != "" ]; then  # The user wants an incremental/referential backup
   # Find out if this is an incremental or diffential backup by the presence of the reference_archive.txt file
   if [ -f "$DESTINATION/$REFERENCE_ARCHIVE/reference_archive.txt" ];then
      echo "Creating an incremental backup on the virtual-tape named: $CURRENT_ARCHIVE"  
   else
      echo "Creating a differential backup on the virtual-tape named: $CURRENT_ARCHIVE"
   fi
   echo "The reference virtual-tape is named: $REFERENCE_ARCHIVE"
   verbose "The basename of the reference virtual-tape is: $REFERENCE_BASENAME"
else    # The user wants an incremental/differential backup, but the reference virtual-tape does not exist
   echo "Warning: You specified an incremental/differential backup, but the reference virtual-tape does not exist."
   echo "Creating a full-backup on the virtual-tape named: $CURRENT_ARCHIVE"
fi

## Create the backup in the working directory
echo
echo -n "Creating backup with DAR..."
verbose
verbose "$DAR_COMMAND $REFERENCE_BASENAME"
$DAR_COMMAND $REFERENCE_BASENAME
if [ "$?" != "0" ]; then
   echo "ERROR: Error when executing the backup with DAR. The attempted command was... "
   echo "$DAR_COMMAND $REFERENCE_BASENAME"
   exit 1
fi

# If configured, test the new archive for errors before removing old archive.
if [ $TEST_ARCHIVE = "yes" ]; then
   echo
   echo -n "Testing the archive for errors..."
   verbose
   verbose "$DAR_BINARY -t $DESTINATION/$WORK_DIR/$BASENAME --noconf"
   $DAR_BINARY -t $DESTINATION/$WORK_DIR/$BASENAME --noconf
   if [ "$?" != "0" ]; then
      echo "ERROR: Error when testing the archive. The attempted command was... "
      echo "$DAR_BINARY -t $DESTINATION/$WORK_DIR/$BASENAME --noconf"
      exit 1
   fi
fi

# Test to make sure the catalogue file exists if they want us to use it
if [ $USE_DAR_MANAGER = "yes" ] && [ ! -f "$SARAB_ETC/$CATALOGUE" ]; then
   echo 
   echo -n "Creating new catalogue..."
   verbose
   verbose "Catalogue file does not exist, creating it..."
   verbose "$DAR_MANAGER --create $SARAB_ETC/$CATALOGUE"
   $DAR_MANAGER --create "$SARAB_ETC/$CATALOGUE"
   if [ "$?" != "0" ]; then
      echo "ERROR: Failed to create catalogue file. The attempted command was..."
      echo "$DAR_MANAGER --create $SARAB_ETC/$CATALOGUE"
      exit 1
   fi
fi

# Remove the old backup with the same name, if it exists
if [ -d "$DESTINATION/$CURRENT_ARCHIVE/" ]; then
   verbose "Removing old backup files..."
   verbose "rm -f $DESTINATION/$CURRENT_ARCHIVE/*"
   rm -f $DESTINATION/$CURRENT_ARCHIVE/*
   if [ "$?" != "0" ]; then
      echo "ERROR: Error when removing the old archive files. The attempted command was... "
      echo "rm -f $DESTINATION/$CURRENT_ARCHIVE/* "
      exit 1
   fi
   verbose "rmdir $DESTINATION/$CURRENT_ARCHIVE/"
   rmdir $DESTINATION/$CURRENT_ARCHIVE/
   if [ "$?" != "0" ]; then
      echo "ERROR: Error when removing the old archive directory. The attempted command was... "
      echo "rmdir -f $DESTINATION/$CURRENT_ARCHIVE/ "
      exit 1
   fi
   if [ $USE_DAR_MANAGER = "yes" ]; then
    # We need to go through the catalog and delete by path 
    # Since dar_manager reorders its archive numbers after every deletion we have to
    # reexecute after every deletion to get the updated numbers...
    verbose "Searching for existing backups in catalogue to delete..."
    while [ 1 ]; do
        i=$( $DAR_MANAGER -B $SARAB_ETC/$CATALOGUE -l | grep "\W$DESTINATION/$CURRENT_ARCHIVE\W" | cut -f 2)
        if [ -n "$i" ]; then 
           i=$( echo $i | cut -f 1 -d" ")
           verbose "Deleting record number $i from catalogue..."
           verbose "$DAR_MANAGER -B $SARAB_ETC/$CATALOGUE --delete $i"
           $DAR_MANAGER -B $SARAB_ETC/$CATALOGUE --delete $i
           if [ "$?" != "0" ]; then
              echo "ERROR: Failed to delete catalogue record number $i. The attempted command was..."
              echo "$DAR_MANAGER -B $SARAB_ETC/$CATALOGUE --delete $i"
              exit 1
           fi
        else 
           break
        fi
    done
    fi
fi   

# Create an empty directory to move the archive into
verbose "Creating virtual-tape directory... mkdir $DESTINATION/$CURRENT_ARCHIVE/"
mkdir $DESTINATION/$CURRENT_ARCHIVE/

# Move the new backup files to the correct location
verbose "Moving files to the correct virtual-tape..."
verbose "mv $DESTINATION/$WORK_DIR/* $DESTINATION/$CURRENT_ARCHIVE/"
mv $DESTINATION/$WORK_DIR/* $DESTINATION/$CURRENT_ARCHIVE/
if [ "$?" != "0" ]; then
   echo "ERROR: Error when moving the new archive. The attempted command was... "
   echo "mv $DESTINATION/$WORK_DIR/* $DESTINATION/$CURRENT_ARCHIVE/"
   exit 1
fi

# Update catalogue with new file
if [ $USE_DAR_MANAGER = "yes" ]; then
   # Add to the catalog the new backup
   echo
   echo -n "Adding backup to catalogue"
   verbose
   verbose "$DAR_MANAGER -B $SARAB_ETC/$CATALOGUE --add $DESTINATION/$CURRENT_ARCHIVE/$BASENAME"
   $DAR_MANAGER -B $SARAB_ETC/$CATALOGUE --add $DESTINATION/$CURRENT_ARCHIVE/$BASENAME
   if [ "$?" != "0" ]; then
      echo "ERROR: Error when adding file to catalogue. The attempted command was..."
      echo "$DAR_MANAGER -B $SARAB_ETC/$CATALOGUE --add $DESTINATION/$CURRENT_ARCHIVE/$BASENAME"
      exit 1
   fi
fi


# Copy the statically-compiled DAR executable with the arhive
if [ "$COPY_DAR" = "yes" ]; then
   if [ -f "$DAR_STATIC" ]; then
      verbose "Copying the static Dar executable..."
      verbose "cp $DAR_STATIC $DESTINATION/$CURRENT_ARCHIVE/"
      cp $DAR_STATIC $DESTINATION/$CURRENT_ARCHIVE/
      if [ "$?" != "0" ]; then
         echo "ERROR: Error when copying the static Dar executable.  The attempted command was... "
         echo "cp $DAR_STATIC $DESTINATION/$CURRENT_ARCHIVE/"
         exit 1
      fi
   else
      echo
      echo "Warning: SaraB tried to copy the static Dar executable, but could not find it."
      echo "Skipping copy of static Dar executable."
   fi
fi

# Rotate the rotation.schedule file by one line.
echo
echo "Rotating the rotation.schedule file by one line."
rotate
if [ "$?" != 0 ]; then
   echo "ERROR: Error when rotating the rotation.schedule file."
   exit 1
fi

 
# Calling the archive script if asked too
if [ ! -z "$CALL_ARCHIVE_SCRIPT" ]; then
   echo 
   echo "Calling archive script... "
   verbose
   verbose "$ARCHIVE_SCRIPT $DESTINATION/$CURRENT_ARCHIVE $BASENAME"
   $ARCHIVE_SCRIPT $DESTINATION/$CURRENT_ARCHIVE $BASENAME
   
   if [ "$?" != 0 ]; then
      echo
      echo "Error: $ARCHIVE_SCRIPT exited with a non-zero status"
   else 
      echo "$ARCHIVE_SCRIPT exited ok"
   fi
fi


## Record finish time
echo "Backup completed at $(date)"
END_TIME=$(date +'%s')

TOTAL_TIME=$(expr $END_TIME - $START_TIME)

HOURS=$(expr $TOTAL_TIME / 3600)
REMAINDER=$(expr $TOTAL_TIME % 3600)
MINUTES=$(expr $REMAINDER / 60)
SECONDS=$(expr $REMAINDER % 60)

echo "The backup process took $HOURS hours, $MINUTES minutes and $SECONDS seconds."
echo "****************************************************************************"

}

########################################################################
########################################################################
#
#   Begin processing here
#
########################################################################
########################################################################

# Make sure the user has root permissions
if [ "$UID" != "0" ]; then
   echo "You must have root permissions to run this script. Exiting."
   echo
   exit 1
fi

# Check if sarab.conf was given as a parameter
if [ -n "$1" ]; then
   # The configuration file was specified as a parameter
   SARAB_CONF="$1"
else
   # Use the default location of sarab.conf
   SARAB_CONF="/etc/sarab/sarab.conf"
fi

# Make sure the sarab.conf file exists
if [ ! -f "$SARAB_CONF" ]; then
   echo "ERROR: You must have $SARAB_CONF to run SaraB."
   echo
   exit 1
fi

# Read the configuration settings
source "$SARAB_CONF"

REQUIRED_FILES="$DAR_BINARY
$SARAB_ETC/$SARAB_DCF
$SARAB_ETC/$ROTATION_SCHEDULE"

## Exit if dependant files do not exist...
for ITEM in $REQUIRED_FILES
do
   if [ ! -f "$ITEM" ]; then
        echo "ERROR: You must have $ITEM to continue" | tee -a $TEMP_LOG
        MAIN_EXIT=1
   fi
done

## Only run if error was NOT detected above
if [ "$MAIN_EXIT" != "1" ]; then
   ## If run from the command line, always output the results to the console
   if [ -z $PS1 ]; then
      # Process the archive. Display and log all output and errors
      main | tee $TEMP_LOG 2>&1
      MAIN_EXIT=${PIPESTATUS[0]}
   else  
      # If run from cron, do not display any output, but still log everything
      # Process the archive and record all output and errors
      main > $TEMP_LOG 2>&1
      MAIN_EXIT="$?"
   fi
fi

## If notification of success is desired, send notification email
if [ "$EMAIL_SUCCESS" = "yes" ] && [ "$MAIN_EXIT" = "0" ]; then
   cat $TEMP_LOG | mail -s "SaraB: Successful backup for $BASENAME" "$EMAIL_ADDRESS"
fi

## If notification of failure is desired, send notification email
if [ "$EMAIL_FAILURE" = "yes" ] && [ "$MAIN_EXIT" != "0" ]; then
   cat $TEMP_LOG | mail -s "SaraB: Backup failure for $BASENAME" "$EMAIL_ADDRESS"
fi

# Append temporary log to main log file
cat $TEMP_LOG >> $LOG_FILE

# Remove temporary log file
rm $TEMP_LOG

exit 0
